

#include "rt_config.h"
#include "rtmp_igmpSnoop.h"


spinlock_t MulticastFilterTabLock;
static MEMBER_ENTRY freeMemberPool[FREE_MEMBER_POOL_SIZE];
static LIST_HEADER freeEntryList;

static inline void initFreeEntryList(
	IN PLIST_HEADER pList)
{
	int i;

	for (i = 0; i < FREE_MEMBER_POOL_SIZE; i++)
		insertTailList(pList, (PLIST_ENTRY)&freeMemberPool[i]);
	return;
}


static VOID IGMPTableDisplay(
	IN PMULTICAST_FILTER_TABLE pMulticastFilterTable);

static BOOLEAN isIgmpMacAddr(
	IN PUCHAR pMacAddr);

static BOOLEAN isIgmpPkt(
	IN PUCHAR pDstMacAddr,
	IN PUCHAR pIpHeader);

static VOID InsertIgmpMember(
	IN PLIST_HEADER pList,
	IN PUCHAR pAddr);

static VOID DeleteIgmpMember(
	IN PLIST_HEADER pList,
	IN PUCHAR pAddr);

static VOID DeleteIgmpMemberList(
	IN PLIST_HEADER pList);

/*
    ==========================================================================
    Description:
        This routine init the entire IGMP table.
    ==========================================================================
 */
VOID MulticastFilterTableInit(
	IN PMULTICAST_FILTER_TABLE *ppMulticastFilterTable)
{
	// Initialize MAC table and allocate spin lock
	*ppMulticastFilterTable = kmalloc(sizeof(MULTICAST_FILTER_TABLE), MEM_ALLOC_FLAG);
	if (*ppMulticastFilterTable == NULL)
	{
		DBGPRINT(RT_DEBUG_ERROR, "%s unable to alloc memory for Multicase filter table, size=%d\n",
			__FUNCTION__, sizeof(MULTICAST_FILTER_TABLE));
		return;
	}

	NdisZeroMemory(*ppMulticastFilterTable, sizeof(MULTICAST_FILTER_TABLE));
	initList(&freeEntryList);
	initFreeEntryList(&freeEntryList);
	NdisAllocateSpinLock(&MulticastFilterTabLock);
	return;
}

/*
    ==========================================================================
    Description:
        This routine reset the entire IGMP table.
    ==========================================================================
 */
VOID MultiCastFilterTableReset(
	IN PMULTICAST_FILTER_TABLE *ppMulticastFilterTable)
{
	ULONG IrqFlags;

	if(*ppMulticastFilterTable == NULL)
		return;

#ifndef THREAD_ISR
	RTMP_SEM_LOCK(&MulticastFilterTabLock, IrqFlags);
#endif
    
#ifdef THREAD_ISR
	RTMP_IRQ_LOCK(IrqFlags);
#endif
	kfree(*ppMulticastFilterTable);
	*ppMulticastFilterTable = NULL;
#ifdef THREAD_ISR
	RTMP_IRQ_UNLOCK(IrqFlags);
#endif	

#ifndef THREAD_ISR
	RTMP_SEM_UNLOCK(&MulticastFilterTabLock, IrqFlags);
#endif
}

/*
    ==========================================================================
    Description:
        Display all entrys in IGMP table
    ==========================================================================
 */
static VOID IGMPTableDisplay(
	IN PMULTICAST_FILTER_TABLE pMulticastFilterTable)
{
	int i;
	MULTICAST_FILTER_TABLE_ENTRY *pEntry = NULL;
	unsigned long IrqFlags;

	if (pMulticastFilterTable == NULL)
		return;

	// if FULL, return
	if (pMulticastFilterTable->Size == 0)
	{
		DBGPRINT(RT_DEBUG_ERROR, "Table empty.\n");
		return;
	}

	// allocate one MAC entry
	RTMP_SEM_LOCK(&MulticastFilterTabLock, IrqFlags);

	for (i = 0; i< MAX_LEN_OF_MULTICAST_FILTER_TABLE; i++)
	{
		// pick up the first available vacancy
		if (pMulticastFilterTable->Content[i].Valid == TRUE)
		{
			PMEMBER_ENTRY pMemberEntry = NULL;
			pEntry = &pMulticastFilterTable->Content[i];

			DBGPRINT(RT_DEBUG_ERROR, "IF(%s) entry #%d, type=%s, GrpId=(%02x:%02x:%02x:%02x:%02x:%02x) memberCnt=%d\n",
				pEntry->net_dev->name, i, (pEntry->type==0 ? "static":"dynamic"),
				pEntry->Addr[0], pEntry->Addr[1], pEntry->Addr[2],
				pEntry->Addr[3], pEntry->Addr[4], pEntry->Addr[5],
				IgmpMemberCnt(&pEntry->MemberList));

			pMemberEntry = (PMEMBER_ENTRY)pEntry->MemberList.pHead;
			while (pMemberEntry)
			{
				DBGPRINT(RT_DEBUG_ERROR, "member mac=(%02x:%02x:%02x:%02x:%02x:%02x)\n",
					pMemberEntry->Addr[0], pMemberEntry->Addr[1], pMemberEntry->Addr[2],
					pMemberEntry->Addr[3], pMemberEntry->Addr[4], pMemberEntry->Addr[5]);

				pMemberEntry = pMemberEntry->pNext;
			}
		}
	}

	RTMP_SEM_UNLOCK(&MulticastFilterTabLock, IrqFlags);

	return;
}

/*
    ==========================================================================
    Description:
        Add and new entry into MAC table
    ==========================================================================
 */
BOOLEAN MulticastFilterTableInsertEntry(
	IN PMULTICAST_FILTER_TABLE pMulticastFilterTable,
	IN PUCHAR pGrpId,
	IN PUCHAR pMemberAddr,
	IN PNET_DEV dev,
	IN MulticastFilterEntryType type)
{
	UCHAR HashIdx;
	int i;
	MULTICAST_FILTER_TABLE_ENTRY *pEntry = NULL, *pCurrEntry, *pPrevEntry;
	PMEMBER_ENTRY pMemberEntry;
	unsigned long IrqFlags;

	if (pMulticastFilterTable == NULL)
		return FALSE;

	// if FULL, return
	if (pMulticastFilterTable->Size >= MAX_LEN_OF_MULTICAST_FILTER_TABLE)
		return FALSE;

	// check the rule is in table already or not.
	if ((pEntry = MulticastFilterTableLookup(pMulticastFilterTable, pGrpId, dev)))
	{
		// doesn't indicate member mac address.
		if(pMemberAddr == NULL)
			return FALSE;

		pMemberEntry = (PMEMBER_ENTRY)pEntry->MemberList.pHead;
		while (pMemberEntry)
		{
			if (MAC_ADDR_EQUAL(pMemberAddr, pMemberEntry->Addr))
				return FALSE;
			pMemberEntry = pMemberEntry->pNext;
		}
	}

	RTMP_SEM_LOCK(&MulticastFilterTabLock, IrqFlags);

	do
	{
		// the multicast entry already exist but doesn't include the member yet.
		if (pEntry != NULL)
		{
			InsertIgmpMember(&pEntry->MemberList, pMemberAddr);
			break;
		}

		// allocate one MAC entry
		for (i = 0; i < MAX_LEN_OF_MULTICAST_FILTER_TABLE; i++)
		{
			// pick up the first available vacancy
			pEntry = &pMulticastFilterTable->Content[i];
			if ((pEntry->Valid == TRUE) && (pEntry->type == MCAT_FILTER_DYNAMIC)
				&& ((jiffies - pEntry->lastTime) > IGMPMAC_TB_ENTRY_AGEOUT_TIME))
			{
				PMULTICAST_FILTER_TABLE_ENTRY pHashEntry;

				HashIdx = MULTICAST_ADDR_HASH_INDEX(pEntry->Addr);
				pHashEntry = pMulticastFilterTable->Hash[HashIdx];

				if ((pEntry->net_dev == pHashEntry->net_dev)
					&& MAC_ADDR_EQUAL(pEntry->Addr, pHashEntry->Addr))
				{
					pMulticastFilterTable->Hash[HashIdx] = pHashEntry->pNext;
					pMulticastFilterTable->Size --;
					DBGPRINT(RT_DEBUG_TRACE, "MCastFilterTableDeleteEntry 1 - Total= %d\n", pMulticastFilterTable->Size);
				} else
				{
					while (pHashEntry->pNext)
					{
						pPrevEntry = pHashEntry;
						pHashEntry = pHashEntry->pNext;
						if ((pEntry->net_dev == pHashEntry->net_dev)
							&& MAC_ADDR_EQUAL(pEntry->Addr, pHashEntry->Addr))
						{
							pPrevEntry->pNext = pHashEntry->pNext;
							pMulticastFilterTable->Size --;
							DBGPRINT(RT_DEBUG_TRACE, "MCastFilterTableDeleteEntry 2 - Total= %d\n", pMulticastFilterTable->Size);
							break;
						}
					}
				}
				pEntry->Valid = FALSE;
				DeleteIgmpMemberList(&pEntry->MemberList);
			}

			if (pEntry->Valid == FALSE)
			{
				NdisZeroMemory(pEntry, sizeof(MULTICAST_FILTER_TABLE_ENTRY));
				pEntry->Valid = TRUE;

				COPY_MAC_ADDR(pEntry->Addr, pGrpId);
				pEntry->net_dev = dev;
				pEntry->lastTime = jiffies;
				pEntry->type = type;
				initList(&pEntry->MemberList);
				InsertIgmpMember(&pEntry->MemberList, pMemberAddr);

				pMulticastFilterTable->Size ++;

				DBGPRINT(RT_DEBUG_TRACE, "MulticastFilterTableInsertEntry -IF(%s) allocate entry #%d, Total= %d\n", dev->name, i, pMulticastFilterTable->Size);
				break;
			}
		}

		// add this MAC entry into HASH table
		if (pEntry)
		{
			HashIdx = MULTICAST_ADDR_HASH_INDEX(pGrpId);
			if (pMulticastFilterTable->Hash[HashIdx] == NULL)
			{
				pMulticastFilterTable->Hash[HashIdx] = pEntry;
			} else
			{
				pCurrEntry = pMulticastFilterTable->Hash[HashIdx];
				while (pCurrEntry->pNext != NULL)
					pCurrEntry = pCurrEntry->pNext;
				pCurrEntry->pNext = pEntry;
			}
		}
	}while(FALSE);

	RTMP_SEM_UNLOCK(&MulticastFilterTabLock, IrqFlags);

	return TRUE;
}

/*
    ==========================================================================
    Description:
        Delete a specified client from MAC table
    ==========================================================================
 */
BOOLEAN MulticastFilterTableDeleteEntry(
	IN PMULTICAST_FILTER_TABLE pMulticastFilterTable,
	IN PUCHAR pGrpId,
	IN PUCHAR pMemberAddr,
	IN PNET_DEV dev)
{
	USHORT HashIdx;
	MULTICAST_FILTER_TABLE_ENTRY *pEntry, *pPrevEntry;
	unsigned long IrqFlags;

	if (pMulticastFilterTable == NULL)
		return FALSE;

	RTMP_SEM_LOCK(&MulticastFilterTabLock, IrqFlags);

	do
	{
		HashIdx = MULTICAST_ADDR_HASH_INDEX(pGrpId);
		pEntry = pMulticastFilterTable->Hash[HashIdx];

		// check the rule is in table already or not.
		if (pEntry && (pMemberAddr != NULL))
		{
			DeleteIgmpMember(&pEntry->MemberList, pMemberAddr);
			if (IgmpMemberCnt(&pEntry->MemberList) > 0)
				break;
		}

		if (pEntry)
		{
			if ((dev == pEntry->net_dev)
				&& MAC_ADDR_EQUAL(pEntry->Addr, pGrpId))
			{
				pMulticastFilterTable->Hash[HashIdx] = pEntry->pNext;
				DeleteIgmpMemberList(&pEntry->MemberList);
				NdisZeroMemory(pEntry, sizeof(MULTICAST_FILTER_TABLE_ENTRY));
				pMulticastFilterTable->Size --;
				DBGPRINT(RT_DEBUG_TRACE, "MCastFilterTableDeleteEntry 1 - Total= %d\n", pMulticastFilterTable->Size);
			} else
			{
				while (pEntry->pNext)
				{
					pPrevEntry = pEntry;
					pEntry = pEntry->pNext;
					if ((dev == pEntry->net_dev)
						&& MAC_ADDR_EQUAL(pEntry->Addr, pGrpId))
					{
						pPrevEntry->pNext = pEntry->pNext;
						DeleteIgmpMemberList(&pEntry->MemberList);
						NdisZeroMemory(pEntry, sizeof(MULTICAST_FILTER_TABLE_ENTRY));
						pMulticastFilterTable->Size --;
						DBGPRINT(RT_DEBUG_TRACE, "MCastFilterTableDeleteEntry 2 - Total= %d\n", pMulticastFilterTable->Size);
						break;
					}
				}
			}
		}
	} while(FALSE);

	RTMP_SEM_UNLOCK(&MulticastFilterTabLock, IrqFlags);
    
	return TRUE;
}

/*
    ==========================================================================
    Description:
        Look up the MAC address in the IGMP table. Return NULL if not found.
    Return:
        pEntry - pointer to the MAC entry; NULL is not found
    ==========================================================================
*/
PMULTICAST_FILTER_TABLE_ENTRY MulticastFilterTableLookup(
	IN PMULTICAST_FILTER_TABLE pMulticastFilterTable,
	IN PUCHAR pAddr,
	IN PNET_DEV dev)
{
	ULONG HashIdx;
	PMULTICAST_FILTER_TABLE_ENTRY pEntry = NULL, pPrev = NULL;

	if (pMulticastFilterTable == NULL)
		return NULL;

	HashIdx = MULTICAST_ADDR_HASH_INDEX(pAddr);
	pEntry = pPrev = pMulticastFilterTable->Hash[HashIdx];

	while (pEntry && pEntry->Valid)
	{
		if ((pEntry->net_dev ==  dev)
			&& MAC_ADDR_EQUAL(pEntry->Addr, pAddr))
		{
			pEntry->lastTime = jiffies;
			break;
		}
		else
		{
			if ((pEntry->Valid == TRUE) && (pEntry->type == MCAT_FILTER_DYNAMIC)
				&& ((jiffies - pEntry->lastTime) > IGMPMAC_TB_ENTRY_AGEOUT_TIME))
			{
				// Remove the aged entry
				if (pEntry == pMulticastFilterTable->Hash[HashIdx])
				{
					pMulticastFilterTable->Hash[HashIdx] = pEntry->pNext;
					pPrev = pMulticastFilterTable->Hash[HashIdx];
				}
				else 
				{	
					pPrev->pNext = pEntry->pNext;
				}

				NdisZeroMemory(pEntry, sizeof(MULTICAST_FILTER_TABLE_ENTRY));
				pMulticastFilterTable->Size --;
				pEntry = (pPrev == NULL ? NULL: pPrev->pNext);
				DBGPRINT(RT_DEBUG_TRACE, "MCastFilterTableDeleteEntry 2 - Total= %d\n", pMulticastFilterTable->Size);
			}
			else
			{
				pPrev = pEntry;
				pEntry = pEntry->pNext;
			}
		}
	}

	return pEntry;
}

VOID IGMPSnooping(
	IN PMULTICAST_FILTER_TABLE pMulticastFilterTable,
	IN PUCHAR pDstMacAddr,
	IN PUCHAR pSrcMacAddr,
	IN PUCHAR pIpHeader,
	IN PNET_DEV pDev)
{
	INT i;
	INT IpHeaderLen;
	UCHAR GroupType;
	UINT16 numOfGroup;
	UCHAR IgmpVerType;
	PUCHAR pIgmpHeader;
	PUCHAR pGroup;
	UCHAR AuxDataLen;
	UINT16 numOfSources;
	PUCHAR pGroupIpAddr;
	UCHAR GroupMacAddr[6];
	PUCHAR pGroupMacAddr = (PUCHAR)&GroupMacAddr;

	if(isIgmpPkt(pDstMacAddr, pIpHeader))
	{
		IpHeaderLen = (*(pIpHeader + 2) & 0x0f) * 4;
		pIgmpHeader = pIpHeader + 2 + IpHeaderLen;
		IgmpVerType = (UCHAR)(*(pIgmpHeader));

		DBGPRINT(RT_DEBUG_TRACE, "IGMP type=%0x\n", IgmpVerType);

		switch(IgmpVerType)
		{
		case IGMP_V1_MEMBERSHIP_REPORT: // IGMP version 1 membership report.
		case IGMP_V2_MEMBERSHIP_REPORT: // IGMP version 2 membership report.
			pGroupIpAddr = (PUCHAR)(pIgmpHeader + 4);
			ConvertMulticastIP2MAC(pGroupIpAddr, (PUCHAR *)&pGroupMacAddr);
			DBGPRINT(RT_DEBUG_TRACE, "IGMP Group=%02x:%02x:%02x:%02x:%02x:%02x\n",
				GroupMacAddr[0], GroupMacAddr[1], GroupMacAddr[2], GroupMacAddr[3], GroupMacAddr[4], GroupMacAddr[5]);
			MulticastFilterTableInsertEntry(pMulticastFilterTable, GroupMacAddr, pSrcMacAddr, pDev, MCAT_FILTER_DYNAMIC);
			break;

		case IGMP_LEAVE_GROUP: // IGMP version 1 and version 2 leave group.
			pGroupIpAddr = (PUCHAR)(pIgmpHeader + 4);
			ConvertMulticastIP2MAC(pGroupIpAddr, (PUCHAR *)&pGroupMacAddr);
			DBGPRINT(RT_DEBUG_TRACE, "IGMP Group=%02x:%02x:%02x:%02x:%02x:%02x\n",
				GroupMacAddr[0], GroupMacAddr[1], GroupMacAddr[2], GroupMacAddr[3], GroupMacAddr[4], GroupMacAddr[5]);
			MulticastFilterTableDeleteEntry(pMulticastFilterTable, GroupMacAddr, pSrcMacAddr, pDev);
			break;

		case IGMP_V3_MEMBERSHIP_REPORT: // IGMP version 3 membership report.
			numOfGroup = ntohs(*((UINT16 *)(pIgmpHeader + 6)));
			pGroup = (PUCHAR)(pIgmpHeader + 8);
			for (i=0; i < numOfGroup; i++)
			{
				GroupType = (UCHAR)(*pGroup);
				AuxDataLen = (UCHAR)(*(pGroup + 1));
				numOfSources = ntohs(*((UINT16 *)(pGroup + 2)));
				pGroupIpAddr = (PUCHAR)(pGroup + 4);
				DBGPRINT(RT_DEBUG_TRACE, "IGMPv3 Type=%d, ADL=%d, numOfSource=%d\n", GroupType, AuxDataLen, numOfSources);
				ConvertMulticastIP2MAC(pGroupIpAddr, (PUCHAR *)&pGroupMacAddr);
				DBGPRINT(RT_DEBUG_TRACE, "IGMP Group=%02x:%02x:%02x:%02x:%02x:%02x\n",
					GroupMacAddr[0], GroupMacAddr[1], GroupMacAddr[2], GroupMacAddr[3], GroupMacAddr[4], GroupMacAddr[5]);
				if((GroupType == MODE_IS_EXCLUDE) || (GroupType == CHANGE_TO_EXCLUDE_MODE))
					MulticastFilterTableInsertEntry(pMulticastFilterTable, GroupMacAddr, pSrcMacAddr, pDev, MCAT_FILTER_DYNAMIC);
				if((GroupType == MODE_IS_INCLUDE) || (GroupType == CHANGE_TO_INCLUDE_MODE))
					MulticastFilterTableDeleteEntry(pMulticastFilterTable, GroupMacAddr, pSrcMacAddr, pDev);
				pGroup += (8 + (numOfSources * 4) + AuxDataLen);
			}
			break;

		default:
			DBGPRINT(RT_DEBUG_TRACE, "unknow IGMP Type=%d\n", IgmpVerType);
			break;
		}
	}

	return;
}

#if 0 // Doesn't be used.
BOOLEAN MulticastFilterApply(
	IN PMULTICAST_FILTER_TABLE pMulticastFilterTable,
	IN PUCHAR pDstMacAddr,
	IN PNET_DEV pDev)
{
	if(isIgmpMacAddr(pDstMacAddr))
		return FALSE;

	return MulticastFilterTableLookup(pMulticastFilterTable, pDstMacAddr, pDev) ? FALSE : TRUE;
}
#endif

INT Set_IgmpSn_Enable_Proc(
	IN PRTMP_ADAPTER pAd,
	IN PUCHAR arg)
{
	UINT Enable;
	PNET_DEV pDev;

	pDev = (pAd->IoctlIF == MAIN_MBSSID) ? (pAd->net_dev) : (pAd->PortCfg.MBSSID[pAd->IoctlIF].MSSIDDev);
	Enable = simple_strtol(arg, 0, 10);

	pAd->PortCfg.MBSSID[pAd->IoctlIF].IgmpSnoopEnable = (BOOLEAN)(Enable == 0 ? 0 : 1);
	DBGPRINT(RT_DEBUG_TRACE, "%s::(%s) %s\n", __FUNCTION__, pDev->name, Enable == TRUE ? "Enable IGMP Snooping":"Disable IGMP Snooping");

	return TRUE;
}

INT Set_IgmpSn_Rate_Proc(
 	IN PRTMP_ADAPTER pAd, 
 	IN PUCHAR arg)
{
	UINT rate;

	rate = simple_strtol(arg, 0, 10);

	switch (pAd->PortCfg.PhyMode)
	{
		case PHY_11A: // A only
			switch (rate)
			{
				case 1: pAd->IgmpGroupTxRate = RATE_6; break; // 6M
				case 2:	pAd->IgmpGroupTxRate = RATE_9; break; // 9M
				case 3: pAd->IgmpGroupTxRate = RATE_12; break; // 12M
				case 4: pAd->IgmpGroupTxRate = RATE_18; break; // 18M
				case 5: pAd->IgmpGroupTxRate = RATE_24; break; // 24M
				case 6: pAd->IgmpGroupTxRate = RATE_36; break; // 36M
				case 7: pAd->IgmpGroupTxRate = RATE_48; break; // 48M
				case 8: pAd->IgmpGroupTxRate = RATE_54; break; // 54M
				default:
					DBGPRINT(RT_DEBUG_ERROR, "Unrecognize Rate.\n");
					break;
			}
			break;

		case PHY_11BG_MIXED: // B/G Mixed
		case PHY_11B: // B only
		case PHY_11G: // G Only
		default:
			switch (rate)
			{
				case 1: pAd->IgmpGroupTxRate = RATE_1; break; // 1M
				case 2: pAd->IgmpGroupTxRate = RATE_2; break; // 2M
				case 3: pAd->IgmpGroupTxRate = RATE_5_5; break; // 5.5M
				case 4: pAd->IgmpGroupTxRate = RATE_11; break; // 11M
				case 5: pAd->IgmpGroupTxRate = RATE_6; break; // 6M
				case 6: pAd->IgmpGroupTxRate = RATE_9; break; // 9M
				case 7: pAd->IgmpGroupTxRate = RATE_12; break; // 12M
				case 8:	pAd->IgmpGroupTxRate = RATE_18; break; // 18M
				case 9: pAd->IgmpGroupTxRate = RATE_24; break; // 24M
				case 10: pAd->IgmpGroupTxRate = RATE_36; break; // 36M
				case 11: pAd->IgmpGroupTxRate = RATE_48; break; // 48M
				case 12: pAd->IgmpGroupTxRate = RATE_54; break; // 54M
				default:
					DBGPRINT(RT_DEBUG_ERROR, "Unrecognize Rate.\n");
					break;
			}
			break;
	}

	DBGPRINT(RT_DEBUG_TRACE, "%s:: IGMP Group TxRate=%d\n", __FUNCTION__, pAd->IgmpGroupTxRate);

	return TRUE;
}

INT Set_IgmpSn_AddEntry_Proc(
	IN PRTMP_ADAPTER pAd, 
	IN PUCHAR arg)
{
	INT i;
	BOOLEAN bGroupId = 1;
	PUCHAR value;
	PUCHAR thisChar;
	UCHAR IpAddr[4];
	UCHAR Addr[ETH_LENGTH_OF_ADDRESS];
	UCHAR GroupId[ETH_LENGTH_OF_ADDRESS];
	PUCHAR *pAddr = (PUCHAR *)&Addr;
	PNET_DEV pDev;

	pDev = (pAd->IoctlIF == MAIN_MBSSID) ? (pAd->net_dev) : (pAd->PortCfg.MBSSID[pAd->IoctlIF].MSSIDDev);

	while ((thisChar = strsep((char **)&arg, "-")) != NULL)
	{
		// refuse the Member if it's not a MAC address.
		if((bGroupId == 0) && (strlen(thisChar) != 17))
			continue;

		if(strlen(thisChar) == 17)  //Mac address acceptable format 01:02:03:04:05:06 length 17
		{
			for (i=0, value = rstrtok(thisChar,":"); value; value = rstrtok(NULL,":"))
			{
				if((strlen(value) != 2) || (!isxdigit(*value)) || (!isxdigit(*(value+1))) ) 
					return FALSE;  //Invalid

				AtoH(value, &Addr[i++], 2);
			}

			if(i != 6)
				return FALSE;  //Invalid
		}
		else
		{
			for (i=0, value = rstrtok(thisChar,"."); value; value = rstrtok(NULL,".")) 
			{
				if((strlen(value) > 0) && (strlen(value) <= 3)) 
				{
					int ii;
					for(ii=0; ii<strlen(value); ii++)
						if (!isxdigit(*(value + ii)))
							return FALSE;
				}
				else
					return FALSE;  //Invalid

				IpAddr[i] = (UCHAR)simple_strtol(value, NULL, 10);
				i++;
			}

			if(i != 4)
				return FALSE;  //Invalid

			ConvertMulticastIP2MAC(IpAddr, (PUCHAR *)&pAddr);
		}

		if(bGroupId == 1)
			COPY_MAC_ADDR(GroupId, Addr);

		// Group-Id must be a MCAST address.
		if((bGroupId == 1) && IS_MULTICAST_MAC_ADDR(Addr))
			MulticastFilterTableInsertEntry(pAd->pMulticastFilterTable, GroupId, NULL, pDev, MCAT_FILTER_STATIC);
		// Group-Member must be a UCAST address.
		else if ((bGroupId == 0) && !IS_MULTICAST_MAC_ADDR(Addr))
			MulticastFilterTableInsertEntry(pAd->pMulticastFilterTable, GroupId, Addr, pDev, MCAT_FILTER_STATIC);
		else
		{
			DBGPRINT(RT_DEBUG_TRACE, "%s (%2X:%2X:%2X:%2X:%2X:%2X) is not a acceptable address.\n",
				__FUNCTION__, Addr[0], Addr[1], Addr[2], Addr[3], Addr[4], Addr[5]);
			return FALSE;
		}

		bGroupId = 0;
		DBGPRINT(RT_DEBUG_TRACE, "%s (%2X:%2X:%2X:%2X:%2X:%2X)\n",
			__FUNCTION__, Addr[0], Addr[1], Addr[2], Addr[3], Addr[4], Addr[5]);

	}

	return TRUE;
}

INT Set_IgmpSn_DelEntry_Proc(
	IN PRTMP_ADAPTER pAd, 
	IN PUCHAR arg)
{
	INT i, memberCnt = 0;
	BOOLEAN bGroupId = 1;
	PUCHAR value;
	PUCHAR thisChar;
	UCHAR IpAddr[4];
	UCHAR Addr[ETH_LENGTH_OF_ADDRESS];
	UCHAR GroupId[ETH_LENGTH_OF_ADDRESS];
	PUCHAR *pAddr = (PUCHAR *)&Addr;
	PNET_DEV pDev;

	pDev = (pAd->IoctlIF == MAIN_MBSSID) ? (pAd->net_dev) : (pAd->PortCfg.MBSSID[pAd->IoctlIF].MSSIDDev);

	while ((thisChar = strsep((char **)&arg, "-")) != NULL)
	{
		// refuse the Member if it's not a MAC address.
		if((bGroupId == 0) && (strlen(thisChar) != 17))
			continue;

		if(strlen(thisChar) == 17)  //Mac address acceptable format 01:02:03:04:05:06 length 17
		{
			for (i=0, value = rstrtok(thisChar,":"); value; value = rstrtok(NULL,":")) 
			{
				if((strlen(value) != 2) || (!isxdigit(*value)) || (!isxdigit(*(value+1))) ) 
					return FALSE;  //Invalid

				AtoH(value, &Addr[i++], 2);
			}

			if(i != 6)
				return FALSE;  //Invalid
		}
		else
		{
			for (i=0, value = rstrtok(thisChar,"."); value; value = rstrtok(NULL,".")) 
			{
				if((strlen(value) > 0) && (strlen(value) <= 3)) 
				{
					int ii;
					for(ii=0; ii<strlen(value); ii++)
						if (!isxdigit(*(value + ii)))
							return FALSE;
				}
				else
					return FALSE;  //Invalid

				IpAddr[i] = (UCHAR)simple_strtol(value, NULL, 10);
				i++;
			}

			if(i != 4)
				return FALSE;  //Invalid

			ConvertMulticastIP2MAC(IpAddr, (PUCHAR *)&pAddr);
		}

		if(bGroupId == 1)
			COPY_MAC_ADDR(GroupId, Addr);
		else
			memberCnt++;

		if (memberCnt > 0 )
			MulticastFilterTableDeleteEntry(pAd->pMulticastFilterTable, (PUCHAR)GroupId, Addr, pDev);

		bGroupId = 0;
	}

	if(memberCnt == 0)
		MulticastFilterTableDeleteEntry(pAd->pMulticastFilterTable, (PUCHAR)GroupId, NULL, pDev);

	DBGPRINT(RT_DEBUG_TRACE, "%s (%2X:%2X:%2X:%2X:%2X:%2X)\n",
		__FUNCTION__, Addr[0], Addr[1], Addr[2], Addr[3], Addr[4], Addr[5]);

	return TRUE;
}

INT Set_IgmpSn_TabDisplay_Proc(
	IN PRTMP_ADAPTER pAd, 
	IN PUCHAR arg)
{
	IGMPTableDisplay(pAd->pMulticastFilterTable);
	return TRUE;
}

VOID ConvertMulticastIP2MAC(
	IN PUCHAR pIpAddr,
	IN PUCHAR *ppMacAddr)
{
	if(pIpAddr == NULL)
		return;

	if(ppMacAddr == NULL || *ppMacAddr == NULL)
		return;

	memset(*ppMacAddr, 0, ETH_LENGTH_OF_ADDRESS);
	*(*ppMacAddr) = 0x01;
	*(*ppMacAddr + 1) = 0x00;
	*(*ppMacAddr + 2) = 0x5e;
	*(*ppMacAddr + 3) = pIpAddr[1] & 0x7f;
	*(*ppMacAddr + 4) = pIpAddr[2];
	*(*ppMacAddr + 5) = pIpAddr[3];
	return;
}

static BOOLEAN isIgmpMacAddr(
	IN PUCHAR pMacAddr)
{
	if((pMacAddr[0] == 0x01)
		&& (pMacAddr[1] == 0x00)
		&& (pMacAddr[2] == 0x5e))
		return TRUE;
	return FALSE;
}

static BOOLEAN isIgmpPkt(
	IN PUCHAR pDstMacAddr,
	IN PUCHAR pIpHeader)
{
	UINT16 IpProtocol = ntohs(*((UINT16 *)(pIpHeader)));
	UCHAR IgmpProtocol;

	if(!isIgmpMacAddr(pDstMacAddr))
		return FALSE;

	if(IpProtocol == ETH_P_IP)
	{
		IgmpProtocol = (UCHAR)*(pIpHeader + 11);
		if(IgmpProtocol == IGMP_PROTOCOL_DESCRIPTOR)
				return TRUE;
	}

	return FALSE;
}

static VOID InsertIgmpMember(
	IN PLIST_HEADER pList,
	IN PUCHAR pAddr)
{
	PMEMBER_ENTRY pMemberEntry;

	if((pAddr == NULL) || (pList == NULL))
		return;

	if((pMemberEntry = (PMEMBER_ENTRY)removeHeadList(&freeEntryList)) != NULL)
	{
		NdisZeroMemory(pMemberEntry, sizeof(MEMBER_ENTRY));
		COPY_MAC_ADDR(pMemberEntry->Addr, pAddr);
		insertTailList(pList, (PLIST_ENTRY)pMemberEntry);

		DBGPRINT(RT_DEBUG_TRACE, "%s mac=(%02x:%02x:%02x:%02x:%02x:%02x)\n",
			__FUNCTION__, pMemberEntry->Addr[0], pMemberEntry->Addr[1], pMemberEntry->Addr[2],
			pMemberEntry->Addr[3], pMemberEntry->Addr[4], pMemberEntry->Addr[5]);
	}
	return;
}

static VOID DeleteIgmpMember(
	IN PLIST_HEADER pList,
	IN PUCHAR pAddr)
{
	PMEMBER_ENTRY pCurEntry;

	if((pAddr == NULL) || (pList == NULL) || (pList->pHead == NULL))
		return;

	pCurEntry = (PMEMBER_ENTRY)pList->pHead;
	while (pCurEntry)
	{
		if(MAC_ADDR_EQUAL(pAddr, pCurEntry->Addr))
		{
			delEntryList(pList, (PLIST_ENTRY)pCurEntry);
			insertTailList(&freeEntryList, (PLIST_ENTRY)pCurEntry);
			break;
		}
		pCurEntry = pCurEntry->pNext;
	}

	return;
}

static VOID DeleteIgmpMemberList(
	IN PLIST_HEADER pList)
{
	PMEMBER_ENTRY pCurEntry, pPrvEntry;

	if((pList == NULL) || (pList->pHead == NULL))
		return;

	pPrvEntry = pCurEntry = (PMEMBER_ENTRY)pList->pHead;
	while (pCurEntry)
	{
		delEntryList(pList, (PLIST_ENTRY)pCurEntry);
		pPrvEntry = pCurEntry;
		pCurEntry = pCurEntry->pNext;
		insertTailList(&freeEntryList, (PLIST_ENTRY)pPrvEntry);
	}

	initList(pList);
	return;
}


UCHAR IgmpMemberCnt(
	IN PLIST_HEADER pList)
{
	if(pList == NULL)
		return 0;

	return getListSize(pList);
}

VOID IgmpGroupDelMembers(
	IN PMULTICAST_FILTER_TABLE pMulticastFilterTable,
	IN PUCHAR pMemberAddr,
	IN PNET_DEV pDev)
{
	INT i;
	MULTICAST_FILTER_TABLE_ENTRY *pEntry = NULL;
	
	for (i = 0; i < MAX_LEN_OF_MULTICAST_FILTER_TABLE; i++)
	{
		// pick up the first available vacancy
		pEntry = &pMulticastFilterTable->Content[i];
		if (pEntry->Valid == TRUE)
			MulticastFilterTableDeleteEntry(pMulticastFilterTable, pEntry->Addr, pMemberAddr, pDev);
	}
}

